import { Service } from "../service"
import { MemoryLockService } from "./memoryLockService"
import { MongodbService } from "./mongodbService"
import * as mongodb from "mongodb"
import * as Leaf from "leaf-ts"
import * as UUID from "uuid"

export type Callback<T = any, E extends Error = Error> = (e?: E, result?: T) => void
export const Errors = Leaf.ErrorDoc.build({
    UnknownError: {},
    NotFound: {},
})

export class TTLSessionStorage extends Service {
    name = "TTLSessionStorage"
    dependencies = ["MongodbService"]
    services: {
        MongodbService: MongodbService
    }

    constructor() {
        super()
    }
    initialize() {
        this.ttlCollection = this.services.MongodbService.db.collection(this.collectionName)
        this.ttlCollection.createIndex({
            expire: 1
        }, {
                expireAfterSeconds: 0
            })
        this.ttlCollection.createIndex({
            id: 1
        }, {
                unique: true
            })
    }
    private collectionName = "ttl"
    private ttlCollection: mongodb.Collection
    createTTLSession(option: {
        data: any
        ttl: number
    }, callback: Callback<string>) {
        let expire = new Date(Date.now() + option.ttl)
        let id = UUID.v4()
        let session: TTLSessionData = {
            id,
            data: option.data,
            expire
        }
        this.ttlCollection.insertOne(session, (err) => {
            callback(err, id)
        })
    }
    updateTTLSession(option: {
        id: string
        data: any
        refreshTTL?: number
    }, callback: Callback<TTLSessionData>) {
        let updates = {
            data: option.data
        }
        if (typeof option.refreshTTL == "number") {
            updates["expire"] = new Date(Date.now() + option.refreshTTL)
        }
        this.ttlCollection.findOneAndUpdate({ id: option.id }, {
            $set: updates
        }, (err, result) => {
            if (err) {
                callback(err)
                return
            }
            if (result.ok !== 1) {
                callback(new Errors.UnknownError("Fail to find and replace TTLSession", { via: result.lastErrorObject }))
                return
            }
            callback(null, result.value)
        })
    }
    getTTLSessionDataById<T>(option: { id: string }, callback: Callback<T>) {
        this.ttlCollection.findOne({ id: option.id }, (err, result: TTLSessionData) => {
            if (err) {
                callback(err)
                return
            }
            if (!result) {
                callback(new Errors.NotFound())
                return
            }
            let data: T = result.data
            callback(null, data)
        })
    }
}

export interface TTLSessionData {
    id: string
    data: any
    expire: Date
}
